@@ -19,7 +19,9 @@
 ;;;; interfaces to defining macros
 
 ;;; an IR1 transform
-(defstruct (transform (:copier nil))
+(defstruct (transform (:copier nil)
+                      (:predicate nil)
+                      #-sb-xc-host :no-constructor-defun)
   ;; the function type which enables this transform.
   ;;
   ;; (Note that declaring this :TYPE FUN-TYPE probably wouldn't
@@ -31,8 +33,7 @@
   ;; the transformation function. Takes the COMBINATION node and
   ;; returns a lambda expression, or THROWs out.
   (function (missing-arg) :type function)
-  ;; T if we should emit a failure note even if SPEED=INHIBIT-WARNINGS.
-  (important nil :type (member nil :slightly t))
+  (important nil :type boolean)
   ;; A function with NODE as an argument that checks wheteher the
   ;; transform applies in its policy.
   ;; It used to be checked in the FUNCTION body but it would produce
@@ -41,12 +42,19 @@
   ;; or if another transform could be applied with the right policy.
   (policy nil :type (or null function)))
 
+;;; A normal transform inserted before VOP-TRANSFORMs
+(defstruct (before-vop-transform (:copier nil)
+                                 (:predicate nil)
+                                 (:include transform)))
+
 ;;; A transform inserted at the front of fun-info-transforms and stops
 ;;; other from firing if it has a VOP that can do the job.
 (defstruct (vop-transform (:copier nil)
                           (:predicate nil)
                           (:include transform)))
 
+(declaim (freeze-type transform))
+
 (defun transform-note (transform)
   (or #+sb-xc-host (documentation (transform-function transform) 'function)
       #-sb-xc-host (and (fboundp 'sb-pcl::fun-doc)
@@ -61,14 +69,20 @@
 ;;; one with the same type and note.
 ;;; Argument order is: policy constraint, ftype constraint, consequent.
 ;;; (think "qualifiers + specializers -> method")
-(defun %deftransform (name policy type fun &optional (important :slightly))
+(defun %deftransform (name policy type fun &optional (important t)
+                                                     priority)
+  (declare (inline make-transform))
   (let* ((ctype (specifier-type type))
          (info (fun-info-or-lose name))
          (transforms (fun-info-transforms info))
          (old (find-if (lambda (transform)
-                         (and (if (eq important :vop)
-                                  (typep transform 'vop-transform)
-                                  (not (typep transform 'vop-transform)))
+                         (and (case important (eq important :vop)
+                                    (:vop
+                                     (typep transform 'vop-transform))
+                                    (:before-vop
+                                     (typep transform 'before-vop-transform))
+                                    (t
+                                     (not (typep transform '(or vop-transform before-vop-transform)))))
                               (type= (transform-type transform)
                                      ctype)))
                        transforms)))
@@ -76,22 +90,36 @@
            (style-warn 'redefinition-with-deftransform :transform old)
            (setf (transform-function old) fun
                  (transform-policy old) policy)
-           (unless (eq important :vop)
+           (unless (or (eq important :vop)
+                       (eq important :before-vop))
              (setf (transform-important old) important)))
           (t
            ;; Put vop-transform at the front.
-           (if (eq important :vop)
-               (push (make-vop-transform :type ctype :function fun
-                                         :policy policy)
-                     (fun-info-transforms info))
-               (let ((normal (member-if (lambda (transform)
-                                          (not (typep transform 'vop-transform)))
-                                        transforms))
-                     (transform (make-transform :type ctype :function fun
-                                                :important important
-                                                :policy policy)))
-                 (setf (fun-info-transforms info)
-                       (append (ldiff transforms normal) (list* transform normal)))))))
+           (case important
+             (:before-vop
+              (push (make-before-vop-transform :type ctype :function fun
+                                               :policy policy)
+                    (fun-info-transforms info)))
+             (:vop
+              (let ((normal (member-if (lambda (transform)
+                                         (not (eq (type-of transform) 'before-vop-transform)))
+                                       transforms))
+                    (transform (make-vop-transform :type ctype :function fun
+                                                   :policy policy)))
+                (setf (fun-info-transforms info)
+                      (append (ldiff transforms normal) (list* transform normal)))))
+             (t
+              (let ((normal (member-if (lambda (transform)
+                                         (not (typep transform '(or vop-transform
+                                                                 before-vop-transform))))
+                                       transforms))
+                    (transform (make-transform :type ctype :function fun
+                                               :important important
+                                               :policy policy)))
+                (setf (fun-info-transforms info)
+                      (if (eq priority :last)
+                          (append transforms (list transform))
+                          (append (ldiff transforms normal) (list* transform normal)))))))))
     name))
 
 ;;; Make a FUN-INFO structure with the specified type, attributes
@@ -148,14 +176,14 @@
                        (fun-info-read-only-args old-fun-info) read-only))
                 (t
                  (setf (info :function :info name)
-                       (make-fun-info :attributes attributes
-                                      :derive-type derive-type
-                                      :optimizer optimizer
-                                      :result-arg result-arg
-                                      :call-type-deriver call-type-deriver
-                                      :annotation annotation
-                                      :folder folder
-                                      :read-only-args read-only))))
+                       (make-fun-info attributes
+                                      derive-type
+                                      optimizer
+                                      result-arg
+                                      call-type-deriver
+                                      annotation
+                                      folder
+                                      read-only))))
           (if location
               (setf (getf (info :source-location :declaration name) 'defknown)
                     location)
@@ -166,13 +194,6 @@
 ;;; This macro should be the way that all implementation independent
 ;;; information about functions is made known to the compiler.
 ;;;
-;;; FIXME: The comment above suggests that perhaps some of my added
-;;; FTYPE declarations are in poor taste. Should I change my
-;;; declarations, or change the comment, or what?
-;;;
-;;; FIXME: DEFKNOWN is needed only at build-the-system time. Figure
-;;; out some way to keep it from appearing in the target system.
-;;;
 ;;; Declare the function NAME to be a known function. We construct a
 ;;; type specifier for the function by wrapping (FUNCTION ...) around
 ;;; the ARG-TYPES and RESULT-TYPE. ATTRIBUTES is an unevaluated list
@@ -558,44 +579,44 @@
          (type (lvar-fun-type fun))
          (policy (lexenv-policy (node-lexenv call)))
          (args (combination-args call)))
-    (when (fun-type-p type)
-      (flet ((assert-type (arg type &optional set index)
-               (when (cond (index
-                            (assert-array-index-lvar-type arg type policy))
-                           (t
-                            (when set
-                              (add-annotation arg
-                                              (make-lvar-modified-annotation :caller (lvar-fun-name fun))))
-                            (assert-lvar-type arg type policy)))
-                 (unless trusted (reoptimize-lvar arg)))))
-        (let ((required (fun-type-required type)))
-          (when set
-            (assert-type (pop args)
-                         (pop required)))
+    (flet ((assert-type (arg type &optional set index)
+             (when (cond (index
+                          (assert-array-index-lvar-type arg type policy))
+                         (t
+                          (when set
+                            (add-annotation arg
+                                            (make-lvar-modified-annotation :caller (lvar-fun-name fun))))
+                          (assert-lvar-type arg type policy)))
+               (unless trusted (reoptimize-lvar arg)))))
+      (let ((required (fun-type-required type)))
+        (when set
           (assert-type (pop args)
-                       (if row-major-aref
-                           (pop required)
-                           (type-intersection
-                            (pop required)
-                            (let ((rank (length args)))
-                              (when (>= rank array-rank-limit)
-                                (setf (combination-kind call) :error)
-                                (compiler-warn "More subscripts for ~a (~a) than ~a (~a)"
-                                               (combination-fun-debug-name call)
-                                               rank
-                                               'array-rank-limit
-                                               array-rank-limit)
-                                (return-from array-call-type-deriver))
-                              (specifier-type `(array * ,rank)))))
-                       set)
-          (loop for type in required
-                do
-                (assert-type (pop args) type nil (or (not (and set row-major-aref))
-                                                     args)))
-          (loop for type in (fun-type-optional type)
-                do (assert-type (pop args) type nil t))
-          (loop for subscript in args
-                do (assert-type subscript (fun-type-rest type) nil t)))))))
+                       (pop required)))
+        (assert-type (pop args)
+                     (if row-major-aref
+                         (pop required)
+                         (type-intersection
+                          (pop required)
+                          (let ((rank (length args)))
+                            (when (>= rank array-rank-limit)
+                              (setf (combination-kind call) :error)
+                              (compiler-warn "More subscripts for ~a (~a) than ~a (~a)"
+                                             (combination-fun-debug-name call)
+                                             rank
+                                             'array-rank-limit
+                                             array-rank-limit)
+                              (return-from array-call-type-deriver))
+                            (make-array-type (make-list rank :initial-element '*)
+                                             :element-type *wild-type*))))
+                     set)
+        (loop for type in required
+              do
+              (assert-type (pop args) type nil (or (not (and set row-major-aref))
+                                                   args)))
+        (loop for type in (fun-type-optional type)
+              do (assert-type (pop args) type nil t))
+        (loop for subscript in args
+              do (assert-type subscript (fun-type-rest type) nil t))))))
 
 (defun append-call-type-deriver (call trusted)
   (let* ((policy (lexenv-policy (node-lexenv call)))
