@@ -62,15 +62,18 @@
                    :qword))
          (ea (ea (- (* slot n-word-bytes) lowtag) ptr)))
     (aver (eq size :qword))
-    (cond ((typep value '(and integer (not (signed-byte 32))))
+    (cond ((or (typep value '(and integer (not (signed-byte 32))))
+               (nil-relative-p value))
            (cond (temp
-                  (inst mov temp value)
+                  (if (nil-relative-p value)
+                      (move-immediate temp value)
+                      (inst mov temp value))
                   (inst mov ea temp)
                   ;; uhh, why does this clause return TEMP but the
                   ;; T clause returns nothing in particular?
                   temp)
                  (t
-                  (bug "need temp reg for STOREW of oversized immediate operand"))))
+                  (bug "need temp reg for STOREW of immediate operand"))))
           (t
            (inst mov :qword ea value)))))
 
@@ -84,8 +87,21 @@
 ;;;; macros to generate useful values
 
 (defmacro load-symbol (reg symbol)
-  `(inst mov ,reg (+ nil-value ,(static-symbol-offset symbol))))
+  (cond (symbol `(inst lea ,reg (ea ,(static-symbol-offset symbol) null-tn)))
+        (t `(inst mov ,reg null-tn))))
 
+;;; Access a thread slot at a fixed index. If GPR-TN is provided,
+;;; then it points to 'struct thread', which is relevant only if #+gs-seg.
+(defun thread-slot-ea (slot-index &optional gpr-tn)
+  (declare (type (signed-byte 16) slot-index)) ; arbitrary
+  (if gpr-tn
+      (ea (ash slot-index word-shift) gpr-tn)
+      ;; Otherwise do something depending on #[-+]gs-seg
+      (let (#+gs-seg (thread-tn nil))
+        (ea thread-segment-reg (ash slot-index word-shift) thread-tn))))
+
+;;; Similar to thread-slot-ea, but INDEX in this case does not signify the Nth slot {0,1,2,..}
+;;; but rather the displacement into the thread's storage, added to the thread base address.
 (defun thread-tls-ea (index)
   #+gs-seg (ea :gs index) ; INDEX is either a DISP or a BASE of the EA
   ;; Whether index is an an integer or a register, the EA constructor
@@ -98,6 +114,12 @@
   ;; explicit displacement of 0.  Using INDEX as base avoids the extra byte.
   #-gs-seg (ea index thread-tn))
 
+(defmacro static-constant-ea (name &optional (extra 0)) ; EXTRA is for byte-sized access
+  (declare (notinline position))
+  ;; 6 = number of words between native ptr to NIL-as-list and 1st trailer constant
+  (let ((n (+ 6 (position name +static-space-trailer-constants+))))
+    `(ea (+ ,extra ,(- (ash n word-shift) list-pointer-lowtag)) null-tn)))
+
 ;;; assert that alloc-region->free_pointer and ->end_addr can be accessed
 ;;; using a single byte displacement from thread-tn
 (eval-when (:compile-toplevel)
@@ -105,16 +127,6 @@
   (aver (<= (1+ thread-mixed-tlab-slot) 15))
   (aver (<= (1+ thread-cons-tlab-slot) 15)))
 
-;;; Access a thread slot at a fixed index. If GPR-TN is provided,
-;;; then it points to 'struct thread', which is relevant only if
-;;; #+gs-seg.
-(defun thread-slot-ea (slot-index &optional gpr-tn)
-  (if gpr-tn
-      (ea (ash slot-index word-shift) gpr-tn)
-      ;; Otherwise do something depending on #[-+]gs-seg
-      (let (#+gs-seg (thread-tn nil))
-        (ea thread-segment-reg (ash slot-index word-shift) thread-tn))))
-
 (defmacro load-tl-symbol-value (reg symbol)
   `(inst mov ,reg (thread-slot-ea ,(symbol-thread-slot symbol))))
 
@@ -178,7 +190,7 @@
   ;; straight-line code, e.g. (LIST (LIST X Y) (LIST Z W)) should emit 1 safepoint
   ;; not 3, even if we consider it 3 separate pointer bumps.
   ;; (Ideally we'd only do 1 pointer bump, but that's a separate issue)
-  (inst test :byte rax-tn (ea -8 gc-card-table-reg-tn)))
+  (inst test :byte rax-tn (ea nil-static-space-end-offs null-tn)))
 
 (macrolet ((pa-bits-ea ()
              `(thread-slot-ea thread-pseudo-atomic-bits-slot
@@ -205,18 +217,51 @@
       #-int1-breakpoints (inst break pending-interrupt-trap))
       OUT)))
 
-;;; This macro is purposely unhygienic with respect to THREAD-TN,
-;;; which is either a global symbol macro, or a LET-bound variable,
-;;; depending on #+gs-seg.
-(defmacro pseudo-atomic ((&key ((:thread-tn thread)) elide-if (default-exit t))
+(defmacro define-allocator (name &body body &aux (g (cdr (assoc :generator body))))
+  `(define-vop ,name
+     #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
+     ,@(remove :generator body :key 'car)
+     (:node-var node)
+     (:generator ,(car g) ; cost
+       (macrolet
+           ((instrument-alloc (&rest args) `(emit-instrument-alloc node thread-tn ,@args))
+            (allocation (&rest args) `(emit-allocation node thread-tn ,@args))
+            (alloc-other (&rest args) `(emit-alloc-other node thread-tn ,@args)))
+         (assemble () ,@(cdr g)))))) ; forms
+
+(defmacro pseudo-atomic ((&key elide-if (default-exit t))
                          &body forms)
-  (declare (ignorable thread))
   `(macrolet ((exit-pseudo-atomic () '(emit-end-pseudo-atomic)))
      (unless ,elide-if
        (emit-begin-pseudo-atomic))
      (assemble () ,@forms)
      (when (and ,default-exit (not ,elide-if))
        (exit-pseudo-atomic))))
+
+;;; For now, ALLOCATING is synonymous with PSEUDO-ATOMIC, however there are
+;;; a few distinguishing factors rendering it not exactly the same:
+;;; 1. It could automatically insert a binding of a thread temp register, should
+;;;    we decide that R13 is not permanently wired to the current thread.
+;;; 2. It will have a yielding behavior that is neither like sb-safepoint
+;;;    nor the current pseudo-atomic, when new-and-improved yieldpoint patches
+;;;    (work in progress) are completed.
+;;; 3. It separates pseudo-atomicity for the purpose of object creation from
+;;;    any other reason which it might be required. Consider that for allocation
+;;;    a behavior might be: if there is room in the TLAB, then fulfill the request;
+;;;    otherwise, induce GC to run immediately and then resume the request.
+;;; [Like sb-safepoint, yieldpoints avoid using flag bits in the thread struct,
+;;; but unlike sb-safepoint, ALLOCATING will not emit a trapping instruction
+;;; at the end of every consing operation. Instead, if and only if it calls
+;;; into C due to TLAB overflow might it yield. This makes sense because we
+;;; give GC the opportunity to run exactly there if the region is exhausted,
+;;; or else there will necessarily be a later yieldpoint based on control flow.
+;;; Note that this is not simply a re-statement of point 3 above which suggests
+;;; that it _will_ cause GC to run; but in general ALLOCATION _may_ let GC run.
+;;; Whereas pseudo-atomic never permits GC to run.
+;;; Also note that "yieldpoints" and "safepoints" are more-or-less the same,
+;;; but a new approach to whatever the concept is demands a different name]
+(defmacro allocating (options &body body)
+  `(pseudo-atomic ,options ,@body))
 
 ;;;; indexed references
 
