@@ -55,30 +55,16 @@
 (defun make-dfun-lambda-list (nargs applyp)
   (let ((required (make-dfun-required-args nargs)))
     (if applyp
-        (nconc required
-               ;; Use &MORE arguments to avoid consing up an &REST list
-               ;; that we might not need at all. See MAKE-EMF-CALL and
-               ;; INVOKE-EFFECTIVE-METHOD-FUNCTION for the other
-               ;; pieces.
-               '(&more .dfun-more-context. .dfun-more-count.))
+        (nconc required '(&rest .rest.))
         required)))
 
 (defun make-dlap-lambda-list (nargs applyp)
   (let ((required (make-dfun-required-args nargs)))
-    ;; Return the full lambda list, the required arguments, a form
-    ;; that will generate a rest-list, and a list of the &MORE
-    ;; parameters used.
-    ;; Beware of deep voodoo! The DEFKNOWN for %LISTIFY-REST-ARGS says that its
-    ;; second argument is INDEX, but the THE form below is "weaker" on account
-    ;; of the vop operand restrictions or something that I don't understand.
-    ;; Which is to say, PCL compilation reliably broke when changed to INDEX.
     (if applyp
-        (values (sys-tlab-append required '(&more .more-context. .more-count.))
+        (values (sys-tlab-append required '(&rest .rest.))
                 required
-                '((sb-c:%listify-rest-args
-                   .more-context. (the (and unsigned-byte fixnum)
-                                    .more-count.)))
-                '(.more-context. .more-count.))
+                '((sb-c::%rest-list .rest.))
+                '(.rest.))
         (values required required nil nil))))
 
 (defun make-emf-call (nargs applyp fn-variable &optional emf-type)
@@ -93,14 +79,10 @@
        ;; the :REST-ARG version or the :MORE-ARG version depending on
        ;; the type of the EMF.
        :rest-arg ,(if applyp
-                      ;; Creates a list from the &MORE arguments.
-                      '((sb-c:%listify-rest-args ; See above re. voodoo
-                         .dfun-more-context.
-                         (the (and unsigned-byte fixnum)
-                           .dfun-more-count.)))
+                      '((sb-c::%rest-list .rest.))
                       nil)
        :more-arg ,(when applyp
-                    '(.dfun-more-context. .dfun-more-count.)))))
+                    '(.rest.)))))
 
 (defun make-fast-method-call-lambda-list (nargs applyp)
   (list* '.pv. '.next-method-call. (make-dfun-lambda-list nargs applyp)))
@@ -256,8 +238,8 @@
                  `((let ((value ,read-form))
                      (return-from access (not (unbound-marker-p value))))))
                 (:makunbound
-                 `(progn (setf ,read-form +slot-unbound+)
-                         ,instance))
+                 `((progn (setf ,read-form +slot-unbound+)
+                          ,instance)))
                 (:writer
                  `((return-from access (setf ,read-form ,(car arglist)))))))
           (funcall miss-fn ,@arglist))))))
@@ -282,11 +264,6 @@
       (:makunbound `(progn (setf ,read-form +slot-unbound+) ,(car arglist)))
       (:writer `(setf ,read-form ,(car arglist))))))
 
-(defmacro emit-reader/writer-macro (reader/writer 1-or-2-class class-slot-p)
-  (let ((*precompiling-lap* t))
-    (values
-     (emit-reader/writer reader/writer 1-or-2-class class-slot-p))))
-
 ;; If CACHED-INDEX-P is false, then the slot location is a constant and
 ;; the cache holds layouts eligible to use that index.
 ;; If true, then the cache is a map of layout -> index.
@@ -312,20 +289,9 @@
                     (when cached-index-p 'index)
                     (unless class-slot-p '(slots)))))))
 
-(defmacro emit-one-or-n-index-reader/writer-macro
-    (reader/writer cached-index-p class-slot-p)
-  (let ((*precompiling-lap* t))
-    (values
-     (emit-one-or-n-index-reader/writer reader/writer
-                                        cached-index-p
-                                        class-slot-p))))
-
 (defun emit-miss (miss-fn args applyp)
   (if applyp
-      `(multiple-value-call ,miss-fn ,@args
-                            (sb-c:%more-arg-values .more-context.
-                                                    0
-                                                    .more-count.))
+      `(apply ,miss-fn ,@args .rest.)
       `(funcall ,miss-fn ,@args)))
 
 ;; (cache-emf, return-value):
@@ -400,14 +366,14 @@
      ;; instance-slots-layout instead of for-std-class-p, as if there
      ;; are no layouts there are no slots to worry about.
      (with-unique-names (wrapper)
-       `(cond ((std-instance-p ,argument)
+       `(cond ((%instancep ,argument)
                ,(if slots-var
                     `(let ((,wrapper (%instance-layout ,argument)))
                        (when (layout-for-pcl-obj-p ,wrapper)
                          (setq ,slots-var (std-instance-slots ,argument)))
                        ,wrapper)
                     `(%instance-layout ,argument)))
-              ((fsc-instance-p ,argument)
+              ((function-with-layout-p ,argument)
                ,(if slots-var
                     `(let ((,wrapper (%fun-layout ,argument)))
                        (when (layout-for-pcl-obj-p ,wrapper)
